/*
 * Graph topology
 */
/*
 * Vertices in strip are numbered as follows:
 *
 *     ___ X-2 ___ X-1 ___ 0   ___ 1   ___ 2   ___ 3   ___ 4   __
 *         Y-1     Y-1      0       0       0       0       0
 *        /  \    /  \    /  \    /  \    /  \    /  \    /  \
 *       /    \  /    \  /    \  /    \  /    \  /    \  /    \
 *     X-3 ___ X-2 ___ X-1 ___ 0   ___ 1   ___ 2   ___  3  ___ 4
 *     Y-2     Y-2     Y-2      1       1       1       1       1
 *       \    /  \    /  \    /  \    /  \    /  \    /  \    /
 *        \  /    \  /    \  /    \  /    \  /    \  /    \  /
 *     ___ X-3 ___ X-2 ___ X-1 ___ 0   ___ 1   ___ 2   ___ 3   __
 *         Y-3     Y-3     Y-3      2       2       2       2
 *
 *       .   .   .   .   .   .   .   .   .   .   .   .   .   .   .
 *
 *      X-4 ___ X-3 ___ X-2 ___ X-1 ___ 0   ___ 1   ___ 2   ___ 3
 *       1       1       1       1      Y-2     Y-2     Y-2     Y-2
 *        \    /  \    /  \    /  \    /  \    /  \    /  \    /
 *         \  /    \  /    \  /    \  /    \  /    \  /    \  /
 *      ___ X-4 ___ X-3 ___ X-2 ___ X-1 ___ 0   ___ 1   ___ 2   __
 *           0       0       0       0      Y-1     Y-1     Y-1
 *
 * Node x,y for
 *   0 <= x < X     x = distance along
 *   0 <= y < Y     y = distance across
 *
 * Vertices are in reading order from diagram above ie x varies fastest.
 *
 * Y must be even.  The actual location opposite (0,0) is (X-(Y-1)/2,0),
 * and likewise opposite (0,Y-1) is ((Y-1)/2,0).
 *
 * We label edges as follows:
 *
 *		e:	    \2   /1
 *			     \  /
 *			  ___ 0   __
 *			  3    1   0
 *			     /  \
 *			   4/   5\
 */

#ifndef MGRAPH_H
#define MGRAPH_H

#include "common.h"

#define XBITS 3
#define X (1<<XBITS)
#define YBITS 3
#define Y (1<<YBITS)

/* vertex number:   0000 | y     | x
 *                        YBITS   XBITS
 */

#define N (X*Y)
#define XMASK (X-1)
#define YSHIFT XBITS
#define Y1 (1 << YSHIFT)
#define YMASK ((Y-1) << YSHIFT)

#define DIM (N*D3)

#define V6 6

#define FOR_VERTEX(v) \
  for ((v)=0; (v)<N; (v)++)

#define FOR_VPEDGE(v,e) \
  for ((e)=0; (e)<V6; (e)++)

extern int edge_end2(unsigned v1, int e);
#define EDGE_END2 edge_end2

#define FOR_VEDGE_X(v1,e,v2,init,otherwise)	\
  FOR_VPEDGE((v1),(e))				\
    if (((v2)= EDGE_END2((v1),(e)),		\
	 (init),				\
	 (v2)) < 0) { otherwise; } else

#define NOTHING ((void)0)

#define FOR_VEDGE(v1,e,v2)			\
  FOR_VEDGE_X(v1,e,v2,NOTHING,NOTHING)

#define FOR_EDGE(v1,e,v2)			\
  FOR_VERTEX((v1))				\
    FOR_VEDGE((v1),(e),(v2))

#define FOR_RIM_VERTEX(vy,vx,v)					\
  for ((vy)=0; (vy)<Y; (vy)+=Y-1)				\
    for ((vx)=0; (v)= (vy)<<YSHIFT | (vx), (vx)<X; (vx)++)

typedef double Vertices[N][D3];

#endif /*MGRAPH_H*/
